/*
    kirc -- KDE Infrared Remote Control

    Copyright (C) 1998 Moritz Wenk (wenk@mathematik.uni-kl.de)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_IRC
#ifndef LIRCD_DEVICE
#warning You have lirc but no lircd device in /dev.  Defaulting to /dev/lircd.
#define LIRCD_DEVICE "/dev/lircd"
#endif

#include <kirc.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <kdebug.h>
#include <klocale.h>

#include <qpushbutton.h>
#include <qlabel.h>

#include "wintv.h"

extern int h_errno;

#ifdef myDEBUG
#define kDEBUG
//#define kDEBUG_A
#define kircWARN
#endif

// ============================================================================

KIRCWidget::KIRCWidget( KConfig* pconf )
  : aCodeDict(37), aWidgetDict(11)
{
  if (!pconf)
    fatal("kirc: Null KConfig object");
  pConfig = pconf;
  aAvailableId = 1;
  aCodeDict.setAutoDelete(TRUE);
  aWidgetDict.setAutoDelete(TRUE);
  keyUP=KCodeEntry::normal;
  num10=0;

  ircclient = new KIRCClient(LIRCD_DEVICE);

  configMode=false;

  kc= new KIRCConfig( &aCodeDict );
  kc->hide();

  connect( kc, SIGNAL(apply()), this, SLOT( applySLOT() ));
  connect( kc, SIGNAL(defaults()), this, SLOT( defaultsSLOT() ));
  connect( kc, SIGNAL(cancel()), this, SLOT( cancelSLOT() ));

  connect( ircclient, SIGNAL(sendReceivedDataSIGNAL( unsigned int, char *, char * )),
	   this, SLOT(receiveDataSLOT( unsigned int, char *, char * )) );

  connect( this, SIGNAL( configDataReceivedSIGNAL( unsigned int, char *, char * )),
	   kc, SLOT(receiveCodeSLOT( unsigned int, char *, char * )));

}

// ============================================================================

KIRCWidget::~KIRCWidget()
{
  sync();

  /* remove all widgetEntries */
  aWidgetDict.clear();
  aCodeDict.clear();

  delete ircclient;
  delete kc;
}

// ============================================================================

void KIRCWidget::applySLOT()
{
  kc->hide();
  configMode=false;

  QDictIterator<KCodeEntry> aKeyIt(aCodeDict);
  aKeyIt.toFirst();
  while ( aKeyIt.current() ) {
    aKeyIt.current()->aCurrentKey=aKeyIt.current()->aConfigKey;
    ++aKeyIt;
  }

  sync();
}

// ============================================================================

void KIRCWidget::defaultsSLOT()
{
  kc->hide();
  configMode=false;
}

// ============================================================================

void KIRCWidget::cancelSLOT()
{
  kc->hide();
  configMode=false;
}

// =================================================================

void KIRCWidget::receiveDataSLOT( unsigned int count, char * button, char * remote_name )
{

  if (configMode) {
    emit configDataReceivedSIGNAL( count, button, remote_name );
    return;
  }

  // find pressed button in button dictionary
  QDictIterator<KCodeEntry> aKeyIt(aCodeDict);
  aKeyIt.toFirst();
  while ( aKeyIt.current() && !(aKeyIt.current()->aCurrentKey == button) )
    ++aKeyIt;

  // button was found in dict, send signal
  if ( aKeyIt.current() && aKeyIt.current()->isEnabled ) {

    if ( aKeyIt.current()->typeOfButton == KCodeEntry::once && count > 0 ) return;

    KCodeEntry *kecode= aKeyIt.current();

    switch ( kecode->typeOfCode ) {
    case KCodeEntry::twoDigits:
    case KCodeEntry::up10:
    case KCodeEntry::up20:
    case KCodeEntry::up30:
      keyUP= kecode->typeOfCode;
      num10=0;
      emit specialKeyReceivedSIGNAL( kecode->typeOfCode, 0 );
      return;

    case KCodeEntry::number:
      switch ( keyUP ) {
      case KCodeEntry::twoDigits:
	if ( num10 == 0 ) {
	  num10= QString( aKeyIt.currentKey() ).toInt();
	  return;
	} else {
	  num= QString( aKeyIt.currentKey() ).toInt()-1 + num10 * 10;
	  keyUP=KCodeEntry::normal;
	}
      case KCodeEntry::up10:
	num= QString( aKeyIt.currentKey() ).toInt() + 10;
	keyUP=KCodeEntry::normal;
	break;
      case KCodeEntry::up20:
	num= QString( aKeyIt.currentKey() ).toInt() + 20;
	keyUP=KCodeEntry::normal;
	break;
      case KCodeEntry::up30:
	num= QString( aKeyIt.currentKey() ).toInt() + 30;
	keyUP=KCodeEntry::normal;
break;
      case KCodeEntry::normal:
	num= QString( aKeyIt.currentKey() ).toInt();
	break;
      } // switch keyUP

      emit numKeyReceivedSIGNAL( num );
      break;

    case KCodeEntry::normal:
      QDictIterator<KCodeConnectEntry> kcce(*kecode->pConnectDict);
      kcce.toFirst();

      while ( kcce.current() ) {
	kcce.current()->sig->activate();
	++kcce;
      }

      break;
    } // switch kecode->typeOfCode

  }
}

// ============================================================================

bool KIRCWidget::addCode( QString functionName, QString default_button,
			  KCodeEntry::codeType type, KCodeEntry::buttonType button,
			  bool configurable )
{
  /* search for an existing "functionName" entry */
  KCodeEntry *pEntry = aCodeDict[ functionName ];

  if ( pEntry ) /* already exists : remove it */
    removeCode(functionName);

  pEntry = new KCodeEntry;
  aCodeDict.insert( functionName, pEntry );

  pEntry->aDefaultKey = default_button;
  pEntry->aCurrentKey = default_button;
  pEntry->bConfigurable = configurable;
  pEntry->aAccelId = 0;
  pEntry->typeOfCode= type;
  pEntry->typeOfButton= button;
  pEntry->pConnectDict = NULL;
  pEntry->isEnabled= true;

  if ( !configurable )
    return false;

  /* search an entry in the KConfig object */
  pConfig->setGroup("CONFIG__IRC_buttons");
  if ( !pConfig->hasKey(functionName) )
    return false;

  /* read and recognize the key */
  QStrList slist;
  QString sel;
  int num = pConfig->readListEntry(functionName,slist);
  sel = slist.first();
  pEntry->aCurrentKey = sel;
  if ( num > 1) {
    sel= slist.next();
    pEntry->typeOfButton = sel.toInt() ? KCodeEntry::once : KCodeEntry::often;
  }

  return false;
}

// ============================================================================

void KIRCWidget::removeCode( const QString& functionName )
{
  /* search for an existing "functionName" entry */
  KCodeEntry *pEntry = aCodeDict[ functionName ];

  if ( !pEntry )
    return;

  /* disconnect eventuals connections */
  if ( pEntry->aAccelId ) {
    QDictIterator<KCodeConnectEntry> aConnectIt(*pEntry->pConnectDict);
    aConnectIt.toFirst();
    while ( aConnectIt.current() ) {
      disconnectFunction( aConnectIt.currentKey(), functionName );
      ++aConnectIt;
    }
  }

  aCodeDict.remove( functionName );
}

// ============================================================================

QString KIRCWidget::readCurrentKey( const QString& functionName )
{
  KCodeEntry *pEntry = aCodeDict[ functionName ];

  if ( !pEntry )
    return pEntry->aCurrentKey;

  return NULL;
}

// ============================================================================

QString KIRCWidget::readDefaultKey( const QString& functionName )
{
  KCodeEntry *pEntry = aCodeDict[ functionName ];

  if ( pEntry )
    return pEntry->aDefaultKey;

  return NULL;
}

// =============================================================================

void KIRCWidget::registerWidget( const QString& widgetName, QWidget* currentWidget )
{
  KIRCWidgetEntry *pWEntry = aWidgetDict[ widgetName ];
  if ( pWEntry ) {
    QString str = QString("initKeyWidget : \"%1\" widget already initialized").arg(widgetName);
    kdWarning() << str << endl;
    return;
  }

  pWEntry = new KIRCWidgetEntry;
  aWidgetDict.insert( widgetName, pWEntry );
  pWEntry->sWidgetName= widgetName;
  pWEntry->pWidget = currentWidget;
}
// ========================================================================

void KIRCWidget::connectFunction( const QString& widgetName,
				  const QString& functionName,
				  const QObject* receiver, const char* member,
				  bool activate )
{
QString str;
  /* search for an existing "widgetName" entry */
  KIRCWidgetEntry *pWEntry = aWidgetDict[ widgetName ];
  if ( !pWEntry ) {
    QString str = QString("connectFunction : \"%1\" widget has not been initialized").arg(widgetName);
    kdWarning() << str << endl;
    return;
  }

  /* search for an existing "functionName" entry */
  KCodeEntry *pEntry = aCodeDict[ functionName ];
  if ( !pEntry ) {
    str = QString("connectFunction : \"%1\" function does not exist").arg(functionName);
    kdWarning() << str << endl;
    return;
  }

  if ( pEntry->typeOfCode != KCodeEntry::normal ) {
    str = QString("connectFunction : special key <%1> can not be connected").arg(functionName);
    kdWarning() << str << endl;
    return;
  }

  KCodeConnectEntry *pCEntry;

  /* if the widget is already connected */
  if ( pEntry->pConnectDict ) {
    pCEntry = (*pEntry->pConnectDict)[ widgetName ];
    if ( pCEntry )
      pEntry->pConnectDict->remove( widgetName );
  } else {
    pEntry->pConnectDict = new QDict<KCodeConnectEntry>(13);
    pEntry->pConnectDict->setAutoDelete(TRUE);
    pEntry->aAccelId = aAvailableId;
    aAvailableId++;
  }

  pCEntry = new KCodeConnectEntry;
  pCEntry->pReceiver = (QObject *)receiver;
  pCEntry->sMember = member;
  pCEntry->sig= new QSignal;
  pCEntry->sig->connect( receiver, member);
  pEntry->pConnectDict->insert( widgetName, pCEntry );

  if ( !activate )
    toggleFunction( widgetName, functionName, FALSE );
}

// ============================================================================

void KIRCWidget::toggleFunction( const QString& widgetName,const QString& functionName, bool activate )
{
  QString str;

  /* search for an existing "widgetName" entry */
  KIRCWidgetEntry *pWEntry = aWidgetDict[ widgetName ];
  if ( !pWEntry ) {
    str = QString("(dis)activateFunction : \"%1\" widget has not been initialized").arg(widgetName);
    kdWarning() << str << endl;
    return;
  }

  /* search for an existing "functionName" entry */
  KCodeEntry *pEntry = aCodeDict[ functionName ];
  if ( !pEntry ) {
    str = QString("(dis)activateFunction : \"%1\" function does not exist").arg(functionName );
    kdWarning() << str << endl;
    return;
  }

  pEntry->isEnabled= activate;
}

// ============================================================================

void KIRCWidget::disconnectFunction( const QString& widgetName,
				     const QString& functionName )
{
  /* search for an existing "widgetName" entry */
  KIRCWidgetEntry *pWEntry = aWidgetDict[ widgetName ];
  if ( !pWEntry ) {
    return;
  }

  /* search for an existing "functionName" entry */
  KCodeEntry *pEntry = aCodeDict[ functionName ];
  if ( !pEntry )
    return;
  /* search for a connection */
  KCodeConnectEntry *pCEntry = (*pEntry->pConnectDict)[ widgetName ];
  if ( !pCEntry )
    return;
  pCEntry->sig->disconnect( pCEntry->pReceiver, pCEntry->sMember );
  delete pCEntry->sig;

  //internalDisconnectFunction( widgetName, pWEntry, pEntry, pCEntry );
  // anstatt
  pEntry->pConnectDict->remove( widgetName );

  /* if this was the unique connection of the functionName : delete the
	   pConnectDict */
  if ( pEntry->pConnectDict->count()==0 )
    delete pEntry->pConnectDict;
}

// ============================================================================
/*
void KIRCWidget::internalDisconnectFunction( const QString& widgetName,
					     KIRCWidgetEntry *pWEntry, KCodeEntry *pEntry, KCodeConnectEntry *pCEntry )
{
  pWEntry->deleteItem( pEntry->aAccelId, pCEntry->pReceiver,
		       pCEntry->sMember );
  pEntry->pConnectDict->remove( widgetName );
}
*/
// ============================================================================

void KIRCWidget::disconnectAllFunctions( const QString& widgetName )
{
  QDictIterator<KCodeEntry> aKeyIt(aCodeDict);
  aKeyIt.toFirst();
  while ( aKeyIt.current() ) {
    disconnectFunction( widgetName, aKeyIt.currentKey() );
    ++aKeyIt;
  }
}

// ============================================================================
void KIRCWidget::destroyWidgetEntry( const QString& widgetName )
{
  aWidgetDict.remove( widgetName );
}
// ============================================================================

void KIRCWidget::internalDisconnectAll( const QString& widgetName)
{
  QDictIterator<KCodeEntry> aKeyIt(aCodeDict);
  aKeyIt.toFirst();
  while ( aKeyIt.current() ) {
    if ( aKeyIt.current()->pConnectDict )
      if ( (*aKeyIt.current()->pConnectDict)[ widgetName ] )
	aKeyIt.current()->pConnectDict->remove( widgetName );
    ++aKeyIt;
  }
}

// ============================================================================

void KIRCWidget::sync()
{
  /* write the current values in aCodeDict to KConfig object */
  pConfig->setGroup("CONFIG__IRC_buttons");
  QString dum;
  QDictIterator<KCodeEntry> aKeyIt(aCodeDict);
  aKeyIt.toFirst();
  while ( aKeyIt.current() ) {
    if ( aKeyIt.current()->bConfigurable ) {
      dum  = aKeyIt.current()->aCurrentKey;
      dum += (aKeyIt.current()->typeOfButton==KCodeEntry::once)?",1":",0";
      pConfig->writeEntry( aKeyIt.currentKey(), dum );
    }
    ++aKeyIt;
  }
}

// ============================================================================

bool KIRCWidget::configureCodes( QWidget *parent )
{
  QDictIterator<KCodeEntry> aKeyIt(aCodeDict);
  aKeyIt.toFirst();
  while ( aKeyIt.current() ) {
    aKeyIt.current()->aConfigKey=aKeyIt.current()->aCurrentKey;
    ++aKeyIt;
  }

  kc->newDictSLOT( &aCodeDict );
  kc->show();
  configMode=true;

  return true;
}

// ============================================================================
// ============================================================================

KIRCConfig::KIRCConfig(QDict<KCodeEntry> *_aCodeDict, QWidget * parent = 0, const char * name = 0 )
  : QWidget( parent, name ), aCodeDict(_aCodeDict)
{
  setCaption(i18n("KWinTV - Infrared remote ontrol button bindings"));

  lv= new QListView( this );
  lv->addColumn(i18n("Function"));
  lv->setColumnWidthMode(0,QListView::Manual);
  lv->setColumnWidth(0,150);
  lv->addColumn(i18n("Button on IR control"));
  lv->setColumnWidthMode(1,QListView::Manual);
  lv->setColumnWidth(1,170);
  lv->addColumn(i18n("Mode"));
  lv->setColumnWidthMode(1,QListView::Manual);
  lv->setColumnWidth(2,50);

  lv->setFrameStyle( 51 );
  lv->setLineWidth( 2 );
  lv->setMultiSelection ( false );
  lv->setAllColumnsShowFocus( true );

  newDictSLOT( _aCodeDict );

  resize(400,350);
  setMinimumSize( 400,350 );
  setMaximumSize( 400,350 );

  lv->setGeometry(5,5,390,200);
  connect( lv, SIGNAL( selectionChanged ( QListViewItem * ) ),this, SLOT(selectionChangedSLOT( QListViewItem * )) );

  QFrame* dlgedit_Frame_1;
  dlgedit_Frame_1 = new QFrame( this, "Frame_1" );
  dlgedit_Frame_1->setGeometry( 5, height() - 80 - 40 - 10, 390, 5 );
  dlgedit_Frame_1->setMinimumSize( 10, 10 );
  dlgedit_Frame_1->setMaximumSize( 32767, 32767 );
  dlgedit_Frame_1->setFrameStyle( 52 );

  QLabel *label= new QLabel( "Enter IRC button name or press IRC button:", this );
  label->setGeometry( 20, height() - 90 - 30, 400-40 ,30 );

  le= new QLineEdit( this );
  le->setGeometry( 20, height() - 90 , 400-40 , 30 );
  connect( le, SIGNAL(textChanged(const QString&)), this, SLOT(textChangedSLOT(const QString&)));

  QFrame* dlgedit_Frame_2;
  dlgedit_Frame_2 = new QFrame( this, "Frame_1" );
  dlgedit_Frame_2->setGeometry( 5, height() - 50 , 390, 5 );
  dlgedit_Frame_2->setMinimumSize( 10, 10 );
  dlgedit_Frame_2->setMaximumSize( 32767, 32767 );
  dlgedit_Frame_2->setFrameStyle( 52 );

  QPushButton *b1= new QPushButton( i18n("OK"), this );
  b1->setGeometry(  width()-320, height()-30, 100, 25);
  connect( b1,SIGNAL( clicked () ), this, SLOT( applySLOT() ));

  QPushButton *b2= new QPushButton( i18n("Default"), this );
  b2->setGeometry(  width()-210,height()-30, 100, 25);
  connect( b2,SIGNAL( clicked () ), this, SLOT( defaultsSLOT() ));

  QPushButton *b3= new QPushButton( i18n("Cancel"), this );
  b3->setGeometry( width()-100, height()-30, 90, 25);
  connect( b3,SIGNAL( clicked () ), this, SLOT( cancelSLOT() ));

  newDictInsert=false;
  selChanged=false;
}

// ============================================================================

KIRCConfig::~KIRCConfig()
{
  delete lv;
}

// ============================================================================

void KIRCConfig::newDictSLOT( QDict<KCodeEntry> *_aCodeDict )
{
  char btype[255];

  newDictInsert=true;

  lv->clear();
  aCodeDict= _aCodeDict;
  QDictIterator<KCodeEntry> aKeyIt(*aCodeDict);
  aKeyIt.toFirst();
  while ( aKeyIt.current() ) {
    QString key(aKeyIt.currentKey());
    if ( aKeyIt.current()->typeOfCode == KCodeEntry::number )
      key = i18n("number")+" \""+key+"\"";
    if ( ( aKeyIt.current()->typeOfCode == KCodeEntry::twoDigits ) ||
	 ( aKeyIt.current()->typeOfCode == KCodeEntry::up10 ) ||
	 ( aKeyIt.current()->typeOfCode == KCodeEntry::up20 ) ||
	 ( aKeyIt.current()->typeOfCode == KCodeEntry::up30 ) )
      key = i18n("special key")+" \""+key+"\"";
    strcpy(btype, aKeyIt.current()->typeOfButton == KCodeEntry::once ? QString(i18n("once")).local8Bit() : QString(i18n("repeat")).local8Bit() );
    (void) new QListViewItem( lv,
			      key,
			      aKeyIt.current()->aConfigKey,
			      btype);
    ++aKeyIt;
  }

  newDictInsert = false;

  currentItem = lv->firstChild ();
  lv->setSelected ( currentItem , true );

  aKeyIt.toFirst();
  currentCode = aKeyIt.current();
}

// ============================================================================

void KIRCConfig::applySLOT()
{
  emit apply();
}

// ============================================================================

void KIRCConfig::defaultsSLOT()
{
  currentCode->aConfigKey=currentCode->aCurrentKey;

  QString text(currentCode->aCurrentKey);
  le->setText(text);

  currentItem->setText( 1, currentCode->aCurrentKey);
}

// ============================================================================

void KIRCConfig::cancelSLOT()
{
  emit cancel();
}

// ============================================================================

void KIRCConfig::receiveCodeSLOT( unsigned int count, char * button, char * remote_name )
{
  currentCode->aConfigKey = button;

  le->setText(button);
  currentItem->setText(1, button);
}

// ============================================================================

void KIRCConfig::selectionChangedSLOT( QListViewItem * item )
{
  if (newDictInsert) return;

  currentItem = item;
  QString sitem( item->text(0) );
  if ( -1 != sitem.find('\"')) {
    sitem.remove(0,sitem.find('\"')+1);
    sitem.remove(sitem.find('\"'),sitem.length()-sitem.find('\"'));
    currentCode= (*aCodeDict)[sitem];
  } else {
    currentCode= (*aCodeDict)[item->text(0)];
  }

  selChanged = true;
  le->setText(currentCode->aConfigKey);
  selChanged = false;
}

// ============================================================================

void KIRCConfig::textChangedSLOT( const QString& text )
{
  currentCode->aConfigKey = text;
  currentItem->setText( 1, text );
}

// ============================================================================
// ============================================================================

KIRCClient::KIRCClient( QString _device ) : device(_device)
{
  button = new char[MAX_CODE_LENGTH];
  remote_name = new char[MAX_CODE_LENGTH];
  status = connect_client();
}

bool KIRCClient::connect_client()
{
  struct sockaddr_un saddr;     // lirc uses UNIX sockets
  saddr.sun_family=AF_UNIX;
  strcpy(saddr.sun_path, device.local8Bit());

  sockfd = socket( AF_UNIX, SOCK_STREAM, 0 );
  if ( sockfd == -1 ) {
    kdWarning() << "kwintv: Unable to create UNIX socket!" << endl;
    return false;
  }

  // make the socket non-blocking here, usually using fcntl( O_NONBLOCK )
  if ( ::connect( sockfd, (struct sockaddr *)&saddr, sizeof(saddr) ) == -1 ) {
    return false;
  }

  sn = new QSocketNotifier( sockfd, QSocketNotifier::Read, this );
  QObject::connect( sn, SIGNAL(activated(int)),
		    this, SLOT(receiveDataFromSocketSLOT(int)) );

  if ( !sn->isEnabled() ) {
    kdWarning() << "kwintv: Could not open socket notifier!" << endl;
    return false;
  }

  return true;
}

// ============================================================================

KIRCClient::~KIRCClient()
{
  delete button;
  delete remote_name;

  if ( -1 == ::close(sockfd) )
    kdWarning() << "kwintv: Device " << device << ":  " << strerror(errno) << endl;
}

// ============================================================================

void KIRCClient::receiveDataFromSocketSLOT( int socket )
{
  if ( ::read(socket,buf,MAX_CODE_LENGTH) == -1 )
    kdWarning() << "kwintv: Read error: %s" << strerror(errno) << endl;

  sscanf( buf, "%x %x %s %s", &code, &count, button, remote_name);

  emit sendReceivedDataSIGNAL( count, button, remote_name );
}

// ============================================================================

bool KIRCClient::reConnect( QString _device )
{
  device = _device;
  status = connect_client();
  return status;
}

// ============================================================================
// ============================================================================


#include "kirc.moc"

#endif // HAVE_IRC


