/* 
    KDE Xv interface

    Copyright (C) 2001 George Staikos (staikos@kde.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#include <assert.h>

#include <qwindowdefs.h>
#include <qwidget.h>

#include <kdebug.h>

#include "kxv.h"


#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#ifdef HAVE_LIBXV
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>
#endif


KXv::KXv() {
  xv_adaptors = 0;
  _devs.setAutoDelete(true);
}


KXv::~KXv() {
kdDebug() << "KXv: Close Xv connection." << endl;
_devs.clear();

#ifdef HAVE_LIBXV
  if (xv_adaptors > 0)
    XvFreeAdaptorInfo((XvAdaptorInfo *)xv_adaptor_info);
#endif
}


KXvDeviceList& KXv::devices() {
  return _devs;
}


bool KXv::haveXv() {
#ifndef HAVE_LIBXV
  return false;
#else
  if (Success != XvQueryExtension(qt_xdisplay(),
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL)) 
    return false;

  return true;

#endif
}



KXv* KXv::connect(Drawable d) {
KXv *xvptr;

  xvptr = new KXv;
  if (!xvptr->init(d)) {
    kdDebug() << "KXv: Xv init failed." << endl;
    delete xvptr;
    return NULL;
  }

kdDebug() << "KXv: Xv init completed." << endl;
return xvptr;
}


bool KXv::init(Drawable d) {
#ifndef HAVE_LIBXV
  return false;
#else
  if (Success != XvQueryExtension(qt_xdisplay(), 
                                  &xv_version,
                                  &xv_release,
                                  &xv_request,
                                  &xv_event,
                                  &xv_error)) {
    return false; 
  }

  if (Success != XvQueryAdaptors(qt_xdisplay(),
                                 d,
                                 &xv_adaptors, 
                                 (XvAdaptorInfo **)&xv_adaptor_info)) {
    // Note technically fatal... what to do?
    kdDebug() << "KXv: XvQueryAdaptors failed." << endl;
  }

  XvAdaptorInfo *ai = (XvAdaptorInfo *)xv_adaptor_info;

  for (unsigned int i = 0; i < xv_adaptors; i++) {
    KXvDevice *xvd = new KXvDevice;
    xvd->xv_type = ai[i].type;
    xvd->xv_port = ai[i].base_id;
    xvd->xv_name = ai[i].name;
    xvd->xv_adaptor = i;
    if (ai[i].type & XvInputMask &&
        ai[i].type & XvVideoMask ) {
      kdDebug() << "KXv: Xv VideoMask port " << ai[i].base_id << " was found." 
              << "  Device is: " << ai[i].name << "." << endl;
    }
    if (ai[i].type & XvInputMask &&
        ai[i].type & XvImageMask ) {
      kdDebug() << "KXv: Xv ImageMask port " << ai[i].base_id << " was found." 
              << "  Device is: " << ai[i].name << "." << endl;
    }

    if (xvd->init()) {
      _devs.append(xvd);
    } else {
      delete xvd;
    }
  }


return true;
#endif
}



bool KXvDevice::startVideo(QWidget *w, int dw, int dh) {
  if (!w) return false;
  return startVideo(w->winId(), dw, dh);
}


bool KXvDevice::startVideo(Window w, int dw, int dh) {
#ifndef HAVE_LIBXV
  return false;
#else
  int sx = 0, sy = 0, dx = 0, dy = 0, sw = dw, sh = dh;

  // Must be a video capable device!
  if (!(xv_type & XvVideoMask) || !(xv_type & XvInputMask)) return false;

  if (videoStarted) stopVideo();

  if (xv_port == -1) return false;

  if (-1 != xv_encoding) {
    sw = ((XvEncodingInfo *)xv_encoding_info)[xv_encoding].width;
    sh = ((XvEncodingInfo *)xv_encoding_info)[xv_encoding].height;
  }

  if (!xv_gc) {
    xv_gc = XCreateGC(qt_xdisplay(), w, 0, NULL);
  }

//  ng_ratio_fixup(&dw, &dh, &dx, &dy);

#if 0
kdDebug() << "XvPutVideo: " << qt_xdisplay() 
          << " " << xv_port << " " << w << " " << xv_gc 
          << " " << sx << " " << sy << " " << sw << " " << sh 
          << " " << dx << " " << dy << " " << dw << " " << dh << endl;
#endif
  XvPutVideo(qt_xdisplay(), xv_port, w, xv_gc, sx, sy, sw, sh, dx, dy, dw, dh);

  videoStarted = true;
  videoWindow = w;
  return true;
#endif
}


bool KXvDevice::stopVideo() {
#ifndef HAVE_LIBXV
  return false;
#else
  if (!videoStarted) return true;
  if (xv_port == -1) return false;
  XvStopVideo(qt_xdisplay(), xv_port, videoWindow);
  XFreeGC(qt_xdisplay(), xv_gc);
  xv_gc = NULL;
  videoStarted = false;
  return true;
#endif
}



KXvDevice::KXvDevice() {
  xv_encoding_info = NULL;
  xv_formatvalues = NULL;
  xv_attr = NULL;
  xv_port = -1;
  xv_encoding = -1;
  xv_name = QString::null;
  xv_type = -1;
  xv_adaptor = -1;
  xv_gc = NULL;
  videoStarted = false;
  _attrs.setAutoDelete(true);
}



KXvDevice::~KXvDevice() {
#ifdef HAVE_LIBXV
  _attrs.clear();
  if (videoStarted) stopVideo();
  if (xv_encoding_info)
    XvFreeEncodingInfo((XvEncodingInfo *)xv_encoding_info);
#endif
}



bool KXvDevice::init() {
#ifndef HAVE_LIBXV
  return false;
#else
  assert(xv_port != -1);   // make sure we were prepped by KXv already.
  if (Success != XvQueryEncodings(qt_xdisplay(),
                                  xv_port,
                                  &xv_encodings,
                                  (XvEncodingInfo **)&xv_encoding_info)) {
    kdDebug() << "KXv: Xv QueryEncodings failed.  Dropping Xv support for this device." << endl;
    return false;
  }

  xv_attr = XvQueryPortAttributes(qt_xdisplay(), 
                                  xv_port, 
                                  &xv_encoding_attributes);
  XvAttribute *xvattr = (XvAttribute *)xv_attr;
  // FIXME: No free method for this?
  kdDebug() << "Attributes for port " << xv_port << endl;
  for (int i = 0; i < xv_encoding_attributes; i++) {
      assert(xvattr);
      kdDebug() << "   -> " << xvattr[i].name 
                << ((xvattr[i].flags & XvGettable) ? " get" : "")
                << ((xvattr[i].flags & XvSettable) ? " set" : "")
                << " Range: " << xvattr[i].min_value
                << " -> " << xvattr[i].max_value << endl;

      KXvDeviceAttribute *xvda = new KXvDeviceAttribute;
      xvda->name = xvattr[i].name;
      xvda->min = xvattr[i].min_value;
      xvda->max = xvattr[i].max_value;
      xvda->flags = xvattr[i].flags;
      _attrs.append(xvda);
  }


  XvImageFormatValues  *fo;
  fo = XvListImageFormats(qt_xdisplay(), xv_port, &xv_formats);
  xv_formatvalues = (void *)fo;
  // FIXME: No free method for this?
  kdDebug() << "Image formats for port " << xv_port << endl;
  for (int i = 0; i < xv_formats; i++) {
    assert(fo);
    QString imout;
    imout.sprintf("   0x%x (%c%c%c%c) %s", fo[i].id,
                                           fo[i].id        & 0xff,
                                          (fo[i].id >>  8) & 0xff,
                                          (fo[i].id >> 16) & 0xff,
                                          (fo[i].id >> 24) & 0xff,
                                          ((fo[i].format == XvPacked) ? 
                                          "Packed" : "Planar"));
    kdDebug() << imout << endl;
  }

  
return true;
#endif
}


bool KXvDevice::isVideoSource() {
#ifndef HAVE_LIBXV
  return false;
#else
  if (xv_type & (XvVideoMask|XvInputMask)) 
    return true;
  return false;
#endif
}


const KXvDeviceAttributes& KXvDevice::attributes() {
  return _attrs;
}


bool KXvDevice::getAttributeRange(QString attribute, int *min, int *max) {
  for (KXvDeviceAttribute *at = _attrs.first(); at != NULL; at = _attrs.next()) {
    if (at->name == attribute) {
      if (min) *min = at->min;
      if (max) *max = at->max;
      return true;
    }
  }
return false;
}


bool KXvDevice::getAttribute(QString attribute, int *val) {
#ifndef HAVE_LIBXV
  return false;
#else
  for (KXvDeviceAttribute *at = _attrs.first(); at != NULL; at = _attrs.next()) {
    if (at->name == attribute) {
      if (val)
        XvGetPortAttribute(qt_xdisplay(), xv_port, at->atom(), val);
      return true;
    }
  }
return false;
#endif
}


bool KXvDevice::setAttribute(QString attribute, int val) {
#ifndef HAVE_LIBXV
  return false;
#else
  for (KXvDeviceAttribute *at = _attrs.first(); at != NULL; at = _attrs.next()) {
    if (at->name == attribute) {
      XvSetPortAttribute(qt_xdisplay(), xv_port, at->atom(), val);
      XSync(qt_xdisplay(), False);
      return true;
    }
  }
return false;
#endif
}


const QString& KXvDevice::name() const {
   return xv_name;
}


Atom KXvDeviceAttribute::atom() {
   return XInternAtom(qt_xdisplay(), name.latin1(), False);
}


