/*
    WinTVScreen base class - default implementation

    Copyright (C) 1998 Moritz Wenk (wenk@mathematik.uni-kl.de)
    Copyright (C) 2001 George Staikos (staikos@kde.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <wintv.h>

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <wintvscreen.h>

#include <kiconloader.h>
#include <qpixmap.h>
#include <qpainter.h>
#include <qtextstream.h>

#include <kdebug.h>

// includes needed for X stuff
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/IntrinsicP.h>
#include <X11/ObjectP.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xmd.h>

#if defined(HAVE_LIBXXF86VM)
#include <X11/extensions/xf86vmode.h>
#include <X11/extensions/xf86vmstr.h>
#endif

#include "xhelpers.h"

#ifdef HAVE_LIBXDPMS
/* this library doesn't even have its own prototypes ... */
extern "C" {
#include <X11/extensions/dpms.h>
}
BOOL dpms_on;
#endif

#include <qt_visual.h>


#define OSD


v4lxif *winTVScreenWidget::v4lx = NULL;

//==================== winTVScreenWidget ========================

winTVScreenWidget::winTVScreenWidget( QWidget *_parent, const char *name,
				      int _capture,
				      const char *_bttvdev,
				      int _bpp, int _palette,
				      int _freq, int _input )
  : QWidget( _parent, name ), parent(_parent), devname(_bttvdev),
    defaultInput(_input), defaultFreq(_freq), captureInitIsOn(_capture)
{
    mrm_on = false;

    // do some widget stuff
    setMinimumSize(160, 120);
    setMaximumSize(32000, 32000);

    QColor bkgc(0, 0, 0);
    setBackgroundColor(bkgc);

    title_pm = UserIcon("kwintv_title.png");

    /////////////////////////////////////////////////////////////////////////
    // channel info
    chInfoText = new QLabel( NULL, "label",
			     WStyle_Customize | WStyle_NoBorder | WStyle_Tool );
    chInfoText->setFrameStyle(QFrame::NoFrame);
    //chInfoText->setAutoResize(true);
    chInfoText->hide();

      QFont f("ledfixed");
      chInfoText->setFont( f );

      QColorGroup normal( QColor( (unsigned int)(16777215) ), QColor( (unsigned int)(0) ),
			  QColor( (unsigned int)(16777215) ), QColor( (unsigned int)(16777215) ),
			  QColor( (unsigned int)(16777215) ), QColor( (unsigned int)(16777215) ),
			  QColor( (unsigned int)(16777215) ) );

      QPalette palette( normal, normal, normal );
      chInfoText->setPalette( palette );

    chInfo = OSD_LEVEL_OFF;

    // timer for on screen display
    infoTimer = new QTimer(this);
    connect(infoTimer,SIGNAL(timeout()),this,SLOT(unShowOSDSLOT()));
    // timer for view refresh actions
    viewRefreshTimer = new QTimer(this);
    connect(viewRefreshTimer,SIGNAL(timeout()),this,SLOT(refreshTVActionSLOT()));
    // timer for muting while frequency changes
    muteTimer = new QTimer(this);
    connect(muteTimer,SIGNAL(timeout()),this,SLOT(setMuteStateAfterFChangeSLOT()));
    // timer for audio mode detection
    audioDetectTimer = new QTimer(this);
    connect(audioDetectTimer,SIGNAL(timeout()),this,SLOT(autoDetectAudioModeSLOT()));
    // timer for hiding mouse cursor after some time
    // mouse tracking is set by initialiseTV()
    mouseHideTimer = new QTimer(this);
    connect(mouseHideTimer,SIGNAL(timeout()),this,SLOT(mouseCursorOffTimerSLOT()));

    // install filter for special X events
    installEventFilter( this );

    // we are running
    widget_is_active = 1;

    // set some window flags
    //setWFlags(WRepaintNoErase|WResizeNoErase);

  kdDebug() << "winTVSCreenWidget - base constructor called" << endl;
}


//==============================================================================

void winTVScreenWidget::postInit()
{
  if ( captureInitIsOn ) {
    newTVGeometry( false );
    captureIsOn = 1;
  }
  didViewRefresh = false;
}

//==============================================================================

winTVScreenWidget::~winTVScreenWidget()
{
  if (widget_is_active) {
    widget_is_active = 0;
  }
  kdDebug() << "winTVSCreenWidget - base destructor called" << endl;
}


//==============================================================================

void winTVScreenWidget::moveEvent ( QMoveEvent * theEvent )
{
  doNewConfig = true;
  switch_capture(0);
  //  unShowOSDSLOT();
  newTVGeometry( true );
  switch_capture(1);
}

//==============================================================================

void winTVScreenWidget::resizeEvent( QResizeEvent * theEvent )
{
  doNewConfig = true;
  switch_capture(0);
  //unShowOSDSLOT();
  newTVGeometry( true );
  switch_capture(1);
}

//==============================================================================

void winTVScreenWidget::keyPressEvent ( QKeyEvent *keyE )
{
  emit keyPressedSIGNAL(keyE->key());
  keyE->ignore();
}

// ===============================================================================

void winTVScreenWidget::mouseReleaseEvent ( QMouseEvent * me)
{
  if ((me->button() == RightButton) && (tvPopup != NULL)) {
    tvPopup->popup(mapToGlobal(me->pos()));
  }
  if (me->button() == MidButton) {
    emit toggleTVonlySIGNAL();
  }
}


//==============================================================================

int winTVScreenWidget::initializeTV(WId parentWid)
{
  // init some variables
  doNewConfig = true;
  mapWindow = true;
  allowViewRefresh = true;
  didViewRefresh = true;
  allowViewRefreshVisibility = true;
  didViewMove = false;
  viewRefreshDelay = VIEWREFRESH_DELAY;
  infodelay = INFO_DELAY;
  infoIsVisible = false;

  muteDelay = 300; // Delay after program switching

  cursorIsVisible = 1;
  allowCursorHiding = 1;
  muteIsOn = 0;
  captureIsOn = 0;

  shotnumber = 0;
  title_shown = 0;

  oldxadj = oldyadj = 0;
  nnewclips = noldclips = 0;

  setTVMaxSize();

  return(0);
}


bool winTVScreenWidget::canFullScreen() {
   return !viewRefreshTimer->isActive();
}


//==============================================================================
//---------------------------------- SLOTS -------------------------------------
//==============================================================================

//==============================================================================

void winTVScreenWidget::changeCaptureSLOT(int value)
{
  if ((captureIsOn && value) || (!captureIsOn && !value)) {
     return;
  }

  captureIsOn = !captureIsOn;
  capture(value);
}

//==============================================================================

void winTVScreenWidget::mouseCursorOffTimerSLOT()
{
  if (mouseHideTimer->isActive()) mouseHideTimer->stop();
  orgCursor = cursor();
  setCursor(blankCursor);
  cursorIsVisible = 0;
}

//==============================================================================

void winTVScreenWidget::toggleMouseCursorHidingSLOT()
{
  allowCursorHiding = allowCursorHiding ? 0 : 1;

  if (!cursorIsVisible) {
    if (mouseHideTimer->isActive()) mouseHideTimer->stop();
    cursorIsVisible = 1;
    setCursor(orgCursor);
  }
}


//==============================================================================

void winTVScreenWidget::enableOSDSLOT( int level )
{
  chInfo = level;
}

//==============================================================================

void winTVScreenWidget::unShowOSDSLOT()
{
  if (infoIsVisible) {
    infoTimer->stop();
    infoIsVisible = false;

    //    doNewConfig= true;
    switch_capture(0);
    if (chInfoText->isVisible()) chInfoText->hide();
    switch_capture(1);
    setTVClipTab(getTVClipTab(false));
    //configOverlay();

    //switch_capture(1);
  }
}


//==============================================================================

void winTVScreenWidget::fullScreenMode(bool on, int xp, int yp, int width, int height, int dw, int dh, int screen)
{

  mrm_on = on;
  mrm_x1 = xp;
  mrm_y1 = yp;
  mrm_x2 = width;
  mrm_y2 = height;
  mrm_dw = dw;
  mrm_dh = dh;
  _screen = screen;
  didViewRefresh = false;
  didViewMove = true;

  if (on && XineramaGetScreenForPoint(QCursor::pos().x(), QCursor::pos().y()) == _screen)
    mouseWasIn = true;
  else mouseWasIn = false;
}



//==============================================================================
void winTVScreenWidget::mouseDoubleClickEvent(QMouseEvent *) {
  emit fullScreenRequest();
}



void flip(char *mem, int len)
{
  int i;
  char tmp;

  for (i = 0; i < len; i++) {
    tmp = mem[2];
    mem[2] = mem[0];
    mem[0] = tmp;
    mem += 3;
  }
}


XErrorHandler xerror(Display *, XErrorEvent *event)
{
#ifdef tvDEBUG
  switch (event->error_code){
  case BadWindow:
    warning("screen: X error BadWindow ingnored");
    break;
  case BadMatch:
    warning("screen: X error BadMatch ingnored");
    break;
  case BadColor:
    warning("screen: X error BadColor ingnored");
    break;
  default:
    warning("screen: unknown X error %d ingnored",event->error_code);
  }
#endif
  return 0;
}



#include "wintvscreen.moc"




