/*
    xWinTV using Bt848 frame grabber driver

    Copyright (C) 1998 Moritz Wenk (wenk@mathematik.uni-kl.de)
    Copyright (C) 2001 George Staikos (staikos@kde.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "wintv_defs.h"

#include <wintv.h>
#include <assert.h>

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <linux/errno.h> // for RedHat

#ifdef HAVE_LIBJPEG
  #include <jpeglib.h>
#endif

#include <wintvscreen_v4l.h>


#include <kiconloader.h>
#include <qpixmap.h>
#include <qpainter.h>
#include <qtextstream.h>
#include <qdatetime.h>

#include <kdebug.h>

// includes needed for X stuff
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/IntrinsicP.h>
#include <X11/ObjectP.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xmd.h>

#if defined(HAVE_LIBXXF86VM)
#include <X11/extensions/xf86vmode.h>
#include <X11/extensions/xf86vmstr.h>
#endif

#include <qt_visual.h>

#include "xhelpers.h"

// debug switches
#ifdef myDEBUG
//#define tvDEBUG
//#define X11DEBUG
#define CLIPDEBUG
#include <xeventtype.h>
char * visibility[] = {"Un","Partially","Fully" , NULL };
#endif

#define HEX(N) hex << setw(N) << setfill(int('0'))
#define DEC(N) dec << setw(N) << setfill(int('0'))

extern int errno;  // for RedHat

#define OSD

//==================== WinTVScreen_v4l ======================================

WinTVScreen_v4l::WinTVScreen_v4l( QWidget *_parent, const char *name,
				      int _capture,
				      const char *_bttvdev,
				      int _bpp, int _palette,
				      int _freq, int _input )
  : winTVScreenWidget( _parent, name, _capture, _bttvdev, _bpp, _palette, _freq, _input)
{

    if (!v4lx)
      v4lx = new v4l1if( devname, _bpp, _palette );

    v4lx->setCapture( off );

    // use my XErrorHandler
    oldXErrorHandler = XSetErrorHandler((XErrorHandler) xerror);

    // init the rest
    initializeTV(parent->winId());

    // we are running
    widget_is_active = 1;

    // Init v4l
    v4lxifP(_bttvdev, _bpp, _palette);
}

//==============================================================================

void WinTVScreen_v4l::postInit()
{
  winTVScreenWidget::postInit();
  v4lx->setCapture( captureIsOn );
}

//==============================================================================

WinTVScreen_v4l::~WinTVScreen_v4l()
{
//kdDebug() << "WinTVScreen_v4l destructor called" << endl;
  if (widget_is_active) {
    destroyTV();
  }
}

//==============================================================================
// X11 clipping stuff

void WinTVScreen_v4l::configOverlay()
{
  if ( mapWindow && (newVisibility != VisibilityFullyObscured) ) {

      if ( newVisibility == VisibilityPartiallyObscured ) {
	noldclips = nnewclips;
	nnewclips = getTVClipTab(false);
	if ( (nnewclips != noldclips) || v4lx->changedCapAClip() ) doNewConfig= true;
	//doNewConfig= true;
      } else {
	nnewclips= 0;
      }

#ifdef CLIPDEBUG
      debug("CLIP: configOverlay, new %d, old %d, %s",nnewclips,noldclips,doNewConfig?"new":"ignore");
#endif
      if ( doNewConfig ) {
#ifdef tvDEBUG
	  debug("screen: allowViewRefresh %d, allowViewRefreshVisibility %d",
		allowViewRefresh,allowViewRefreshVisibility);
#endif
	  doNewConfig = false;
	  switch_capture(0);
	  setTVClipTab( nnewclips );
	  switch_capture(1);
	  if ( allowViewRefresh && allowViewRefreshVisibility && !didViewRefresh ) {
	    refreshTV();
	  }
      }

  } else {
#ifdef CLIPDEBUG
      debug("CLIP: configOverlay, off");
#endif
      switch_capture(0);
      setTVClipTab( 0 );
      doNewConfig = false;
      if ( allowViewRefresh && !didViewRefresh ) {
	refreshTV();
      }
  }
}

// X11 events des shell windows
void WinTVScreen_v4l::Shellx11EventSLOT( XEvent *event )
{
//  kdDebug() << "CLIP: x11Event: (TV shell) " << event->type << endl;
  if ( captureIsOn ) {
    switch ( event->type ) {
    case MapNotify:
      doNewConfig = true;
      mapWindow = true;
      configOverlay();
     break;
    case UnmapNotify:
      doNewConfig = true;
      mapWindow = false;
      configOverlay();
     break;
    } // switch
  }
}

// X11 events for the TV window
bool WinTVScreen_v4l::x11Event( XEvent *event )
{
//  if ( event->type != MotionNotify ) kdDebug() << "CLIP: x11Event: (TV win) " << event->type << endl;

  if ( isVisible() && (event->xvisibility.window = (Window) winId()) ) {
    if ( allowCursorHiding && event->type == MotionNotify ) {
      if ( !cursorIsVisible ) {
	setCursor(orgCursor);
	cursorIsVisible = true;
      }

      if (mouseHideTimer->isActive()) mouseHideTimer->stop();

      mouseHideTimer->start(MOUSE_DELAY,true);

#ifdef HAVE_LIBXXF86VM
      XLockDisplay(qt_xdisplay());
      if ( mrm_on ) {
        int new_x1 = mrm_x1+2*mrm_dw;
	int new_y1 = mrm_y1;
	XF86VidModeGetViewPort(qt_xdisplay(), _screen, &mrm_x, &mrm_y);
	if ( mrm_x != new_x1 || mrm_y != new_y1 ) {
           int _x,_y,_w,_h;
           XineramaGetScreenGeometry(_screen, &_x, &_y, &_w, &_h);
           XF86VidModeSetViewPort(qt_xdisplay(), _screen, new_x1, new_y1);
           if (mouseWasIn) {
              int xx,yy,ww,hh;
              QPoint cp = QCursor::pos();

              if (cp.x()-_x <= new_x1) {
                int next = XineramaLeftOf(_screen);
                if (next >= 0) {
                  XineramaGetScreenGeometry(next, &xx, &yy, &ww, &hh);
                  QCursor::setPos(xx+ww-5, cp.y());
                  mouseWasIn = false;
                } else QCursor::setPos(new_x1+mrm_x2/2+_x, new_y1+mrm_y2/2+_y); 
              } else if (cp.x()-_x >= new_x1+mrm_x2) {
                int next = XineramaRightOf(_screen);
                if (next >= 0) {
                  XineramaGetScreenGeometry(next, &xx, &yy, &ww, &hh);
                  QCursor::setPos(xx+5, cp.y());
                  mouseWasIn = false;
                } else QCursor::setPos(new_x1+mrm_x2/2+_x, new_y1+mrm_y2/2+_y); 
              } else if (cp.y()-_y <= new_y1) {
                int next = XineramaAbove(_screen);
                if (next >= 0) {
                  XineramaGetScreenGeometry(next, &xx, &yy, &ww, &hh);
                  QCursor::setPos(cp.x(), yy+hh-5);
                  mouseWasIn = false;
                } else QCursor::setPos(new_x1+mrm_x2/2+_x, new_y1+mrm_y2/2+_y); 
              } else if (cp.y()-_y >= new_y1+mrm_y2) {
                int next = XineramaBelow(_screen);
                if (next >= 0) {
                  XineramaGetScreenGeometry(next, &xx, &yy, &ww, &hh);
                  QCursor::setPos(cp.x(), yy+5);
                  mouseWasIn = false;
                } else QCursor::setPos(new_x1+mrm_x2/2+_x, new_y1+mrm_y2/2+_y); 
              } else QCursor::setPos(new_x1+mrm_x2/2+_x, new_y1+mrm_y2/2+_y);
           } else {
              QCursor::setPos(new_x1+mrm_x2/2+_x, new_y1+mrm_y2/2+_y);
              mouseWasIn = true;
           }
        }
      }
      XUnlockDisplay(qt_xdisplay());
#endif
      return false;
    }

    if ( captureIsOn ) {
      switch ( event->type ) {
      case Expose:
	if ( !event->xexpose.count ) {
	  if (newVisibility == VisibilityUnobscured) {
	    doNewConfig = true;
	    configOverlay();
	  }
	}
	break;
      case VisibilityNotify:
	newVisibility = event->xvisibility.state;
	switch (event->xvisibility.state) {
	case VisibilityFullyObscured:
          if ( !didViewRefresh ) configOverlay();
	  break;
	case VisibilityPartiallyObscured:
	case VisibilityUnobscured:
	  doNewConfig = true;
	  mapWindow = true;
	  if ( didViewRefresh ) {
	    didViewRefresh = false;
	  } else {
	    configOverlay();
	  }
	  break;
	}

//kdDebug() << "CLIP: x11Event: (TV win) did it refresh? " << (didViewRefresh?"y":"n") << endl;
//if ( mapWindow != captureIsOn ) kdDebug() << "CLIP: x11Event: (TV win) call configOverlay()" << endl;
      } // switch (event->type)
    }
  }
  return (false);
}

// X11 events for the root window (desktop)
bool WinTVScreen_v4l::eventFilter( QObject *, QEvent *e )
{

  X11RWEvent *re = (X11RWEvent*)e;

  if ( captureIsOn && isVisible() &&
       (e->type() == XRWEvent) &&
       (re->XEventPtr()->xvisibility.window != (Window) winId()) ) {

//kdDebug() << "CLIP: eating event: " << re->XEventTyp() << " Did refresh? " << (didViewRefresh?"y":"n") << endl;

    switch (re->XEventTyp()) {
    case MapNotify:
    case UnmapNotify:
    case ConfigureNotify:

      if ( !didViewRefresh )
	configOverlay();

     break;
    } // switch

    return true;                        // eat event
  }

  return false;                         // standard event processing
}

//==============================================================================

void WinTVScreen_v4l::paintEvent( QPaintEvent * event)
{
  if ( !captureIsOn ) {
//    kdDebug() << "X11: paintEvent" << endl;
    if ( !title_shown ) {
      QPainter pa(this);
      pa.setClipRect(event->rect());
      pa.drawPixmap((width()/2) - (title_pm.width()/2),
		    (height()/2) - (title_pm.height()/2),
		    title_pm);
    } else {
      //title_shown=1;
      kdDebug() << "Alert:  Please report this to the developer. 4939241" << endl;
    }
  }
}

//==============================================================================

void WinTVScreen_v4l::switch_capture(int on)
{
//kdDebug() << "screen: switch_capture from " << captureIsOn << " to " << on << endl;
  if (on && captureIsOn) {
    if ( isVisible() ) {
      v4lx->setCapture( on );
      captureIsOn = true;
    }
  } else {
    v4lx->setCapture( off );
  }
}

//==============================================================================

void WinTVScreen_v4l::capture(int on)
{
//kdDebug() << "screen: capture from " << captureIsOn << " to " << on << endl;
  if (isVisible()) {
    if (on) {
      captureIsOn = true;
      newTVGeometry( false );
    } else {
      captureIsOn = false;
    }
    v4lx->setCapture( captureIsOn );
  }
}

//==============================================================================

int WinTVScreen_v4l::initializeTV(WId parentWid)
{
  Window rootw;
  Display *disp;
  ushort i;
  int visual_class;
  Visual *myvisual;
  //  char visname[255];

  // get the display
  disp = qt_xdisplay();

  // get the visual and check for doublebuffering
  find_visual(disp,NULL,&myvisual,&visual_class);

  colorMap8 = false;
  // we have to use PseudoColor, i.e. we have only 256 colors
  if (visual_class == PseudoColor) {
    Colormap xcmap;
    int r, rr;
#ifdef tvDEBUG
    debug("screen: using color class PseudoColor");
#endif
    colorMap8 = true;
    for (i = 0; i < 256; i++) {
      xcolors[i].pixel = i;
      xcolors[i].flags = DoRed | DoGreen | DoBlue;
    }

    //XQueryColors(disp, xcolormap, xcolors, 256);
    xcmap = XCreateColormap(disp, DefaultRootWindow(disp),myvisual, AllocAll);
    for ( i = 0; i < 225; i++){ // Set Colors 16 - 240
      r = i%25;
      xcolors[i+16].green = (65535*(i/25)/8);
      xcolors[i+16].red = (65535*(r/5)/4);
      rr = r%5;
      xcolors[i+16].blue = (65535*rr/4);
    }

    XStoreColors(disp, xcmap, xcolors, 256);
    XSetWindowColormap(disp, (Window)parentWid, xcmap);
#ifdef tvDEBUG
    debug("screen: set up colormap for 8 bpp");
#endif
  }

  // set up some X stuff
  rootw = DefaultRootWindow(disp);

  // tv window -- need visibility, structure
  XSelectInput( disp, (Window) winId(), ButtonPressMask      | 
                                        ButtonReleaseMask    |
                                        ExposureMask         | 
                                        PointerMotionMask    | 
                                        VisibilityChangeMask | 
                                        StructureNotifyMask   );

  // root window -- need visibility, sub-, structure
  XSelectInput( disp, rootw, VisibilityChangeMask | 
                             FocusChangeMask      |
                             StructureNotifyMask  | 
                             SubstructureNotifyMask );

  audioVolumeCan = v4lx->capAudioVolume();

  return winTVScreenWidget::initializeTV(parentWid);
}

//==============================================================================

void WinTVScreen_v4l::setTVMaxSize()
{
  unsigned int minw, minh;
  v4lx->capCapSize( &minw, &minh, &max_width, &max_height );
#ifdef tvDEBUG
  debug ("screen: screensize max_x: %d, max_y: %d",max_width,max_height);
#endif
}

//==============================================================================

void WinTVScreen_v4l::destroyTV()
{
  capture(0);

  // reset X stuff
  XSetErrorHandler(oldXErrorHandler);
  if (colorMap8) 
     XFreeColormap(qt_xdisplay(),xcolormap);
  XSelectInput(qt_xdisplay(),DefaultRootWindow(qt_xdisplay()),0);

  // kill the interface
  delete v4lx;

}

//==============================================================================

void WinTVScreen_v4l::newTVGeometry( bool doViewRefresh )
{
  bool checkClipping;
  unsigned int view_width, view_height;
  int xp, yp, oldx, oldy, dx = 0, dy = 0;
  unsigned int neww, newh, oldw, oldh;

  if (!isVisible()) return;

  v4lx->getCapAClip( &oldx, &oldy, &oldw, &oldh );
#ifdef CLIPDEBUG
  debug("CLIP: newTVGeometry values getCapAClip %d %d %d %d",oldx,oldy,oldw,oldh);
#endif

  checkClipping = false;
  didViewMove = false;

  // get new geometry values, need the global position
  QPoint p = mapToGlobal(QPoint(0,0));
  view_height = geometry().height();
  view_width = geometry().width();

  xp = p.x() & ~3;
  yp = p.y() & ~3;

/*
  kdDebug() << "Height: " << view_height
            << " Width: " << view_width
            << " x(): " << p.x()
            << " y(): " << p.y()
            << endl;
*/

  // checkClipping if size changed
  // set a new width
  if (view_width > max_width) {
    dx = (view_width - max_width)/2;
    neww = max_width & ~3;
  } else {
    neww = view_width & ~3;
  }

  if (oldw != neww) checkClipping = true;

  if (oldw > neww) didViewMove = true;

  // set a new height
  if (view_height > max_height) {
    dy = (view_height - max_height)/2;
    newh = max_height & ~3;
  } else {
    newh = view_height & ~3;
  }

  if (oldh != newh) checkClipping = true;

  if (oldh > newh) didViewMove = true;

  if ( (p.x() & ~3) < p.x() ) {
    xp += 4;
    neww -= 4;
#ifdef CLIPDEBUG
    debug("CLIP: xp, neww");
#endif
  }

  if ( (p.y() & ~3) < p.y() ) {
    yp += 4;
    newh -= 4;
#ifdef CLIPDEBUG
    debug("CLIP: yp, newh");
#endif
  }

  // checkClipping if position changed
  if (oldx != xp+dx) {
    didViewMove = true;
    checkClipping = true;
  } else if (oldy != yp+dy) {
    didViewMove = true;
    checkClipping = true;
  }

  if (checkClipping && captureIsOn) {
    setTVClipTab(getTVClipTab(true,xp+dx,yp+dy,neww,newh)); //,(xp+dx)-oldx,(yp+dy)-oldy)
    if ( allowViewRefresh && doViewRefresh ) refreshTV();
  }

}

//==============================================================================

void WinTVScreen_v4l::autoDetectAudioModeSLOT()
{
  v4lx->setAudioMode( 0 );
  if (audioMode != v4lx->getAudioMode()) {
    emit autoDetectAudioModeSIGNAL(v4lx->getAudioMode() | AUDIOMODE_AUTODETECT);
  }
}

//==============================================================================

void WinTVScreen_v4l::refreshTVActionSLOT()
{
  int scr;
  XSetWindowAttributes xswa;
  unsigned long mask;
  Window rwin,rootW;
  Display *disp;

  viewRefreshTimer->stop();

  if ( !didViewRefresh && ((newVisibility != VisibilityUnobscured) || didViewMove) ) {
    disp = qt_xdisplay();
    scr = DefaultScreen(disp);
    rootW = RootWindow(disp, scr);

    xswa.override_redirect = true;
    xswa.backing_store = NotUseful;
    xswa.save_under = false;
    mask = ( CWSaveUnder | CWBackingStore | CWOverrideRedirect );

    if (didViewMove) {
      rwin = XCreateWindow(disp, rootW, 
                           0, 0,
			   kapp->desktop()->width(),kapp->desktop()->height(),
			   /*vbuf.width, vbuf.height,*/
			   0, CopyFromParent, InputOutput, CopyFromParent,
			   mask, &xswa);
      didViewMove=false;
    } else {
      QPoint p = mapToGlobal(QPoint(0,0));
      rwin = XCreateWindow(disp,rootW,
			   p.x()-40, p.y()-40,
			   geometry().width()+80,geometry().height()+80,
			   0, CopyFromParent, InputOutput, CopyFromParent,
			   mask, &xswa);
    }

    XMapWindow(disp, rwin);
    XUnmapWindow(disp, rwin);
    XDestroyWindow(disp, rwin);

    didViewRefresh = true;
  }
}

//==============================================================================

void  WinTVScreen_v4l::refreshTV()
{
  didViewRefresh = false;
  if ( viewRefreshTimer->isActive() ) viewRefreshTimer->stop();
  viewRefreshTimer->start(viewRefreshDelay);
}

//==============================================================================

int WinTVScreen_v4l::getTVClipTab( bool set,
                                   int wx, int wy,
                                   unsigned int ww, unsigned int wh,
                                   int xadj, int yadj )
{
  int nclips;
  int x, y, x2, y2;
  Display *disp;
  XWindowAttributes wts;
  Window parent, win, root, rroot, *children;
  uint nchildren, i;

  if (!isVisible()) return(0);

  v4lx->resetCapAClip();
  if ( !set ) {
    v4lx->getCapAClip( &wx, &wy, &ww, &wh );
  } else { 
    oldxadj = xadj;
    oldyadj = yadj;
  }

  if (wx<0)
    v4lx->addCapAClip(0, 0, (uint)(-wx), wh);
  if (wy<0)
    v4lx->addCapAClip(0, 0, ww, (uint)(-wy));
  if ((wx+ww) > v4lx->getFBwidth())
    v4lx->addCapAClip(v4lx->getFBwidth()-wx, 0, ww, wh);
  if ((wy+wh) > v4lx->getFBheight())
    v4lx->addCapAClip(0, v4lx->getFBheight()-wy, ww, wh);

  disp = qt_xdisplay();
  win = (Window)winId();
  root = DefaultRootWindow(disp);

  for (;;) {
    XQueryTree(disp, win, &rroot, &parent, &children, &nchildren);
    XFree((char *) children);
    //debug("screen: getTVClip: win=0x%x, parent=0x%x\n",me,parent);
    if (root == parent) break;
    win = parent;
  }

  XQueryTree(disp, root, &rroot, &parent, &children, &nchildren);

  for  (i = 0; i < nchildren; i++ )
    if ( children[i] == win )
      break;

  for ( i++; i < nchildren; i++  ) {
    XGetWindowAttributes(disp, children[i], &wts);
    if ( !(wts.map_state & IsViewable) ) continue;

    x = wts.x - wx;
    y = wts.y - wy;
    x2 = x + wts.width + 2*wts.border_width;
    y2 = y + wts.height + 2*wts.border_width;

    if ( (x2 < 0) || (x >= (int)ww) || (y2 < 0) || (y >= (int)wh) ) continue;

    if ( x < 0 ) x = 0;
    if ( y < 0 ) y = 0;
    if ( x2 >= (int)ww ) x2 = ww;
    if ( y2 >= (int)wh ) y2 = wh;

    nclips = v4lx->addCapAClip(x, y, x2, y2, xadj, yadj);
  }

  XFree((char *) children);

  //kdDebug() << "x,y,w,h " << WinTVScreen_v4l::x() << "," << WinTVScreen_v4l::y() << "," << width() << "," << height() << endl;
  //kdDebug() << "setCapAClip " << wx << " " << wy << " " << ww << " " << wh << endl;
  return v4lx->setCapAClip( wx, wy, ww, wh );
}

//==============================================================================

int WinTVScreen_v4l::setTVClipTab( int num )
{
  return v4lx->applyCapAClip( num );
}

//==============================================================================
//---------------------------------- SLOTS -------------------------------------
//==============================================================================

void WinTVScreen_v4l::snapShotSLOT(const char *prefix, int fixed, int width,
                                   int height, int format, int quality)
{
  QString name,ext;
  char *tm;
  int bpp = 3;

  if (captureIsOn) {

    // set image params and capture one
    v4lx->grabSetParams( fixed, &width, &height, VIDEO_PALETTE_RGB24 );
    tm = v4lx->grabCapture( true );

    switch (format) {
    case SNAPSHOT_JPEG:
      ext = ".jpg";
     break;
    case SNAPSHOT_GIF:
      ext = ".gif";
     break;
    case SNAPSHOT_TIFF:
      ext = ".tif";
     break;
    case SNAPSHOT_PNM:
    default:
      ext = ".pnm";
     break;
    }

    // dont overwrite existing files...
    QString otime; 
    QDateTime now = QDateTime::currentDateTime();
    otime.sprintf("-%04d%02d%02d-%02d%02d%02d-",
		  now.date().year(), now.date().month(), now.date().day(),
		  now.time().hour(), now.time().minute(), now.time().second());

    bool file_exists = true;

    while (file_exists) {
      QTextStream oname( name, IO_WriteOnly );
      oname << prefix << otime << HEX(3) << shotnumber << ext << char(0);

      ifstream os_t(name);
      if (!os_t) {
	file_exists = false;
	os_t.close();
      } else {
	shotnumber++;
	os_t.close();
      }
    }

    QTextStream oname( name, IO_WriteOnly );
    oname << prefix << otime << HEX(3) << shotnumber << ext << char(0);

    switch (format) {
#ifdef HAVE_LIBJPEG
    case SNAPSHOT_JPEG:
      {
	int rs,k;
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	FILE *outfile;

	if (!(outfile = fopen((const char*)name, "wb"))){
	  warning("screen: Can't open file %s: %s",(const char*)name,strerror(errno));
	  return;
	}

	flip(tm, width*height);

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);
	jpeg_stdio_dest(&cinfo,outfile);
	cinfo.image_width = width;
	cinfo.image_height = height;
	cinfo.input_components = bpp;
	cinfo.in_color_space = JCS_RGB;

	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo,quality,true);

	jpeg_start_compress(&cinfo,true);

	rs = width*3;
	JSAMPROW rp[height];

	for (k = 0; k < height; k++)
	  rp[k] = (unsigned char *)&tm[rs*k];

	jpeg_write_scanlines(&cinfo,rp,height);

	/* while(cinfo.next_scanline < cinfo.image_height){
	   JSAMPROW rp[1];
	   rp[0]=& mem[rs*cinfo.next_scanline];
	   int sl=jpeg_write_scanlines(&cinfo,rp,1);
	   }
	   */
	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);

	fclose(outfile);
      }
    break;
#endif
    case SNAPSHOT_GIF:
    case SNAPSHOT_TIFF:
    case SNAPSHOT_PNM:
    default:
      {
	ofstream os((const char*)name);
	if (!os) {
	  warning("screen: Can't open file %s: %s",(const char*)name,strerror(errno));
	  return;
	}
	os << "P6\n" << dec << width << endl << dec << height << "\n255\n";

	// FIX ME do we need this??
	flip(tm, width*height);
	os.write(tm,width*height*3);
	os.close();
	chown(name,getuid(),getgid());
      }
    }

#ifdef tvDEBUG
    debug("screen: snap shot %s",(const char*)name);
#endif

    if ( chInfo > OSD_LEVEL_OFF )
      showOSDSLOT((const char*)(QString("wrote ")+name));

  } // captureIsOn
}

//==============================================================================

void WinTVScreen_v4l::changeFrequencySLOT(int value)
{
  if ( defaultFreq != value ) {
    defaultFreq = value;
    v4lx->setFreq( value );
  }
}

//==============================================================================

void WinTVScreen_v4l::changeBrightnessSLOT(int bri)
{
  if (Bright != bri)
    {
      Bright = bri;
      v4lx->setPicBrightness( bri );
    }
}

//==============================================================================

void WinTVScreen_v4l::changeContrastSLOT(int contr)
{
  if (Contrast != contr)
    {
      Contrast = contr;
      v4lx->setPicConstrast( contr );
    }
}

//==============================================================================

void WinTVScreen_v4l::changeColorSLOT(int color)
{
  if (Color != color)
    {
      Color = color;
      v4lx->setPicColor( color );
    }
}

//==============================================================================

void WinTVScreen_v4l::changeHueSLOT(int hue)
{
  if (Hue != hue)
    {
      Hue = hue;
      v4lx->setPicHue( hue );
    }
}

//==============================================================================

void WinTVScreen_v4l::changeSourceSLOT(int s)
{
  // change the input source (Tuner, Video, ... )
  if (defaultInput != s){
    defaultInput = s;
    v4lx->setChannel( s );
  }
}

//==============================================================================

void WinTVScreen_v4l::changeNormSLOT(int n)
{
#ifdef tvDEBUG
  debug("screen: changeNormSLOT %d",n);
#endif
  // change input norm ( PAL, SECAM, NTSC )
  if (n != defaultNorm) {
    v4lx->setChannelNorm( n );
    defaultNorm = n;
    setTVMaxSize();
    newTVGeometry( true );
  }
}

//==============================================================================

void WinTVScreen_v4l::changeChannelValuesSLOT(channel *c, int id)
{
  if (!c) return;

  switch_capture(0);

  if ( !muteIsOn && v4lx->capChannelTuner(c->input) ) {
    // not muted -> mute for some milisecs...
    emit requestMixerMute(true);
    v4lx->setAudioMute(true);
    qApp->processEvents( 250 ); // give mixer time to mute soundcard
    usleep(SCAN_DELAY);
    if ( muteTimer->isActive() ) muteTimer->stop();
    muteTimer->start(muteDelay);
  }

  changeSourceSLOT(c->input);
  changeNormSLOT(c->norm);
  if ( v4lx->capChannelTuner(c->input) ) {
    changeFrequencySLOT(c->frequency);
  }

  changeBrightnessSLOT(c->brightness);
  changeColorSLOT(c->color);
  changeHueSLOT(c->hue);
  changeContrastSLOT(c->contrast);

  switch_capture(1);

  c->id = id;

  if ( chInfo >= OSD_LEVEL_2 ) showOSDSLOT(c);
}

//==============================================================================

void WinTVScreen_v4l::setMuteStateAfterFChangeSLOT()
{
  muteTimer->stop();
  if ( !muteIsOn ) { // not muted -> unmute again
    v4lx->setAudioMute(false);
    emit requestMixerMute(false);
  }
}

//==============================================================================

void WinTVScreen_v4l::toggleAudioMuteSLOT()
{
  muteIsOn = !muteIsOn;

  v4lx->setAudioMute(muteIsOn);

  if (muteIsOn) {
    if ( chInfo > OSD_LEVEL_1 ) showOSDSLOT("sound muted");
  } else {
    if ( chInfo > OSD_LEVEL_1 ) showOSDSLOT("sound unmuted");
  }
}

//==============================================================================

void WinTVScreen_v4l::changeAudioVolumeSLOT(int vl, int vr)
{
}

//==============================================================================

void WinTVScreen_v4l::changeAudioModeSLOT(int mode)
{
  if (mode > 0) {
    if (audioDetectTimer->isActive()) audioDetectTimer->stop();
    audioMode = mode;

    // try to set mode...
    v4lx->setAudioMode( audioMode );
    audioMode = v4lx->getAudioMode();
    emit autoDetectAudioModeSIGNAL(audioMode);

  } else {
    // autodetection, mode==0
    // set up timer
    audioMode=mode;
    if (audioDetectTimer->isActive()) audioDetectTimer->stop();
    audioDetectTimer->start(AUDIOMODE_DELAY);
  }

}


//==============================================================================

void WinTVScreen_v4l::showOSDSLOT(channel *ch)
{
#ifdef OSD
  int wx,wy;
  unsigned int wwidth,wheight;

  if ( !captureIsOn || !isVisible() || (chInfo == OSD_LEVEL_OFF) ) return;

  if (infoTimer->isActive()) infoTimer->stop();
  infoTimer->start(infodelay,true);

  v4lx->getCapAClip( &wx, &wy, &wwidth, &wheight );

  chInfoText->setGeometry(wx+(wwidth/10),wy+wheight-(wheight/10)-10,
			  (wwidth/10 * 8)+7 & ~3,(wheight/10)+7 & ~3);

  QString tout;
  tout.setNum(ch->id + 1);
  tout.insert(0," ");
  tout += " -- ";
  tout += ch->name;
  tout = tout.leftJustify(25,' ');
  tout += QTime::currentTime().toString();
  chInfoText->setText(tout);

  infoIsVisible = true;

  switch_capture(0);
  chInfoText->show();
  switch_capture(1);
#endif
}

//==============================================================================

void WinTVScreen_v4l::showOSDSLOT(const char * text)
{
#ifdef OSD
  int wx,wy;
  unsigned int wwidth,wheight;

  if ( !captureIsOn || !isVisible() || (chInfo == OSD_LEVEL_OFF) ) return;

  if (infoTimer->isActive()) infoTimer->stop();

  v4lx->getCapAClip( &wx, &wy, &wwidth, &wheight );

  chInfoText->setGeometry(wx+(wwidth/10),wy+wheight-(wheight/10)-10,
			  (wwidth/10 * 8)+7 & ~3,(wheight/10)+7 & ~3);

  QString tout(" ");
  tout += text;
  chInfoText->setText(tout);

  infoIsVisible = true;
  switch_capture(0);
  chInfoText->show();
  switch_capture(1);

  infoTimer->start(infodelay,true);

#endif
}


void WinTVScreen_v4l::rawSetFreq(unsigned long freq) {
    v4lx->setFreq(freq);
}


void WinTVScreen_v4l::rawCapTunerRange(unsigned long *fmin, unsigned long *fmax) {
    v4lx->capTunerRange(fmin, fmax);
}


void WinTVScreen_v4l::rawSetChannel(int source) {
    v4lx->setChannel(source);
}


void WinTVScreen_v4l::rawSetChannelNorm(int norm) {
    v4lx->setChannelNorm(norm);
}


unsigned long WinTVScreen_v4l::rawGetTunerSignal() {
    return v4lx->getTunerSignal();
}



#include "wintvscreen_v4l.moc"




